package trask.app.life;					  

import java.awt.*;
import java.util.*;

public class CalculationDisplay extends Container
{
  private boolean useDisjointSubBlocks = false;

  CalculationDetailsDisplay calculationDetails;

  private Label temperatureLabel;
  private Label harmonyLabel;
  Label lifeLabel;
  private Label complexityLabel;

  private int displayPrecision = 2;

  public CalculationDisplay()
  {
    calculationDetails = new CalculationDetailsDisplay();

    temperatureLabel = new Label("", Label.LEFT);
    harmonyLabel = new Label("", Label.LEFT);
    lifeLabel = new Label("", Label.LEFT);
    complexityLabel = new Label("", Label.LEFT);

    setLayout(new GridBagLayout());

    GridBagConstraints constraints = new GridBagConstraints();

    // global constraints
    constraints.gridwidth = 1;
    constraints.gridheight = 1;
    constraints.fill = GridBagConstraints.NONE;

    // name label constraints
    constraints.anchor = GridBagConstraints.EAST;
    constraints.gridx = 0;

    constraints.gridy = 0;
    add(new Label("Temperature:"), constraints);
    constraints.gridy = 1;
    add(new Label("Harmony:"), constraints);
    constraints.gridy = 2;
    add(new Label("Life:"), constraints);
    constraints.gridy = 3;
    add(new Label("Complexity:"), constraints);

    // value label constraints
    constraints.anchor = GridBagConstraints.WEST;
    constraints.fill = GridBagConstraints.HORIZONTAL;
    constraints.gridx = 1;

    constraints.gridy = 0;
    add(temperatureLabel, constraints);
    constraints.gridy = 1;
    add(harmonyLabel, constraints);
    constraints.gridy = 2;
    add(lifeLabel, constraints);
    constraints.gridy = 3;
    add(complexityLabel, constraints);
  }

  // get and set methods

  public void setUseDisjointSubBlocks(boolean useDisjoint)
  {
    useDisjointSubBlocks = useDisjoint;
  }

  public int getDisplayPrecision()
  {
    return displayPrecision;
  }

  public void setDisplayPrecision(int precision)
  {
    displayPrecision = precision;
    calculationDetails.setDisplayPrecision(precision);
  }

  public void setCalculationDetailsVisible(boolean visible)
  {
    calculationDetails.setVisible(visible);
  }

  // ... methods

  public void calculateLife(Color[][] picture, Dimension pictureDim)
  {
    int minimumDim = Math.min(pictureDim.width, pictureDim.height);

    Vector[] subBlockVector = new Vector[minimumDim + 1];
    double[] temperature = new double[minimumDim + 1];
    double[] harmony = new double[minimumDim + 1];

    Cursor parentCursor = null;
    Cursor waitCursor = new Cursor(Cursor.WAIT_CURSOR);
    Container parent = getParent();
    if (parent != null) {
      parentCursor = parent.getCursor();
      parent.setCursor(waitCursor);
    }

    if (useDisjointSubBlocks) {
      subBlockVector[0] = null;
      subBlockVector[1] = null;
      for (int i = 2; i < subBlockVector.length; i++) {
        if (pictureDim.width % i == 0 && pictureDim.height % i == 0) {
          subBlockVector[i] = getDisjointSubBlocks(pictureDim, i);
        } else {
          subBlockVector[i] = null;
        }
      }
    } else {
      // use Overlapping SubBlocks
      subBlockVector[0] = null;
      subBlockVector[1] = null;
      for (int i = 2; i < subBlockVector.length; i++) {
        subBlockVector[i] = getOverlappingSubBlocks(pictureDim, i);
      }
    }

    double totalTemperature = 0;
    int totalNum = 0;
    double averageTemperature;
    for (int i = 2; i < subBlockVector.length; i++) {
      if (subBlockVector[i] != null) {
        temperature[i] = calculateAverageTemperature(picture, subBlockVector[i]);
        totalTemperature += temperature[i];
        totalNum++;
      }
    }
    averageTemperature = totalTemperature / totalNum;

    double totalHarmony = 0;
    totalNum = 0;
    double averageHarmony;
    for (int i = 2; i < subBlockVector.length; i++) {
      if (subBlockVector[i] != null) {
        harmony[i] = calculateAverageHarmony(picture, subBlockVector[i]);
        totalHarmony += harmony[i];
        totalNum++;
      }
    }
    averageHarmony = totalHarmony / totalNum;

    calculationDetails.setDetails(temperature, harmony, subBlockVector);

    double life = averageTemperature * averageHarmony;
    double complexity = averageTemperature * (9 - averageHarmony);

    double tenpow = Math.pow(10, displayPrecision);

    temperatureLabel.setText("" + (Math.round(averageTemperature * tenpow) / tenpow)) ;
    harmonyLabel.setText("" + (Math.round(averageHarmony * tenpow) / tenpow));
    lifeLabel.setText("" + (Math.round(life * tenpow) / tenpow));
    complexityLabel.setText("" + (Math.round(complexity * tenpow) / tenpow));

    temperatureLabel.invalidate();
    harmonyLabel.invalidate();
    lifeLabel.invalidate();
    complexityLabel.invalidate();
    validateTree();

    if (parent != null) {
      parent.setCursor(parentCursor);
    }
  }

  private Vector getDisjointSubBlocks(Dimension pictureDim, int blockSize)
  {
    Vector subBlocks = new Vector();

    for (int i = 0; i < pictureDim.width; i += blockSize) {
      for (int j = 0; j < pictureDim.height; j += blockSize) {
        Rectangle subBlock = new Rectangle(i, j, blockSize, blockSize);
        subBlocks.addElement(subBlock);
      }
    }

    return subBlocks;
  }

  private Vector getOverlappingSubBlocks(Dimension pictureDim, int blockSize)
  {
    Vector subBlocks = new Vector();

    for (int i = 0; i < pictureDim.width - blockSize + 1; i++) {
      for (int j = 0; j < pictureDim.height - blockSize + 1; j++) {
        Rectangle subBlock = new Rectangle(i, j, blockSize, blockSize);
        subBlocks.addElement(subBlock);
      }
    }

    return subBlocks;
  }

  private double calculateAverageTemperature(Color[][] picture, Vector subBlocks)
  {
    double total = 0;

    for (Enumeration e = subBlocks.elements(); e.hasMoreElements(); ) {
      Rectangle subBlock = (Rectangle) e.nextElement();
      total += calculateTemperature(picture, subBlock);
    }

    return total / subBlocks.size();
  }

  private double calculateTemperature(Color[][] picture, Rectangle subBlock)
  {
    Color[] colors = new Color[subBlock.width * subBlock.height];

    for (int i = 0; i < colors.length; i++) {
      colors[i] = picture[subBlock.x + i % subBlock.width]
                         [subBlock.y + i / subBlock.width];
    }

    for (int i = 0; i < colors.length; i++) {
      if (colors[i] == null) continue;
      for (int j = i + 1; j < colors.length; j++) {
        if (colors[j] == colors[i]) colors[j] = null;
      }
    }

    int total = 0;

    for (int i = 0; i < colors.length; i++) {
      if (colors[i] != null) total++;
    }

    return total - 1;
  }

  private double calculateAverageHarmony(Color[][] picture, Vector subBlocks)
  {
    double total = 0;

    for (Enumeration e = subBlocks.elements(); e.hasMoreElements(); ) {
      Rectangle subBlock = (Rectangle) e.nextElement();
      total += calculateElementSymmetry(picture, subBlock);
      total += calculateTranslationalSymmetry(picture, subBlock, subBlocks);
    }

    return total / subBlocks.size();
  }

  private double calculateElementSymmetry(Color[][] picture, Rectangle subBlock)
  {
    Color[][] colorBlock = makeColorBlock(picture, subBlock);

    int total = 0;

    if (xAxisSymmetry(colorBlock, colorBlock)) total++;
    if (yAxisSymmetry(colorBlock, colorBlock)) total++;
    if (posDiagonalSymmetry(colorBlock, colorBlock)) total++;
    if (negDiagonalSymmetry(colorBlock, colorBlock)) total++;
    if (rotationalSymmetry90(colorBlock, colorBlock)) total++;
    if (rotationalSymmetry180(colorBlock, colorBlock)) total++;

    return total;
  }

  private double calculateTranslationalSymmetry(Color[][] picture,
                              Rectangle subBlock, Vector subBlocks)
  {
    boolean simpleSymmetry = false;
    boolean axisSymmetry = false;
    boolean rotationalSymmetry = false;
    int total = 0;

    Color[][] colorBlock = makeColorBlock(picture, subBlock);

    for (Enumeration e = subBlocks.elements(); e.hasMoreElements(); ) {
      Rectangle subBlockCompare = (Rectangle) e.nextElement();
      if (subBlockCompare == subBlock) continue;

      Color[][] colorBlockCompare = makeColorBlock(picture, subBlockCompare);

      if (!simpleSymmetry && simpleSymmetry(colorBlock, colorBlockCompare)) {
        simpleSymmetry = true;
        total++;
      }
      if (!axisSymmetry && (xAxisSymmetry(colorBlock, colorBlockCompare)
          || yAxisSymmetry(colorBlock, colorBlockCompare))) {
        axisSymmetry = true;
        total++;
      }
      if (!rotationalSymmetry && (rotationalSymmetry90(colorBlock, colorBlockCompare)
          || rotationalSymmetry180(colorBlock, colorBlockCompare))) {
        rotationalSymmetry = true;
        total++;
      }

      if (total == 3) break;
    }

    return total;
  }

  private boolean xAxisSymmetry(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    Dimension blockDim = new Dimension(colorBlock.length, colorBlock[0].length);

    for (int i = 0; i < blockDim.width; i++) {
      for (int j = 0; j < blockDim.height; j++) {
        if (colorBlock[i][blockDim.height - 1 - j] != colorBlockCompare[i][j]) {
          return false;
        }
      }
    }
    return true;
  }

  private boolean yAxisSymmetry(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    Dimension blockDim = new Dimension(colorBlock.length, colorBlock[0].length);

    for (int i = 0; i < blockDim.width; i++) {
      for (int j = 0; j < blockDim.height; j++) {
        if (colorBlock[blockDim.width - 1 - i][j] != colorBlockCompare[i][j]) {
          return false;
        }
      }
    }
    return true;
  }

  private boolean posDiagonalSymmetry(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    Dimension blockDim = new Dimension(colorBlock.length, colorBlock[0].length);

    for (int i = 0; i < blockDim.width; i++) {
      for (int j = 0; j < blockDim.height; j++) {
        if (colorBlock[blockDim.height - 1 - j][blockDim.width - 1 - i]
            != colorBlockCompare[i][j]) {
          return false;
        }
      }
    }
    return true;
  }

  private boolean negDiagonalSymmetry(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    Dimension blockDim = new Dimension(colorBlock.length, colorBlock[0].length);

    for (int i = 0; i < blockDim.width; i++) {
      for (int j = 0; j < blockDim.height; j++) {
        if (colorBlock[j][i] != colorBlockCompare[i][j]) {
          return false;
        }
      }
    }
    return true;
  }

  private boolean rotationalSymmetry90(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    return rotationalSymmetry90pos(colorBlock, colorBlockCompare)
           || rotationalSymmetry90neg(colorBlock, colorBlockCompare);
  }

  private boolean rotationalSymmetry90pos(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    Dimension blockDim = new Dimension(colorBlock.length, colorBlock[0].length);

    for (int i = 0; i < blockDim.width; i++) {
      for (int j = 0; j < blockDim.height; j++) {
        if (colorBlock[j][blockDim.width - 1 - i] != colorBlockCompare[i][j]) {
          return false;
        }
      }
    }
    return true;
  }

  private boolean rotationalSymmetry90neg(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    Dimension blockDim = new Dimension(colorBlock.length, colorBlock[0].length);

    for (int i = 0; i < blockDim.width; i++) {
      for (int j = 0; j < blockDim.height; j++) {
        if (colorBlock[blockDim.height - 1 - j][i] != colorBlockCompare[i][j]) {
          return false;
        }
      }
    }
    return true;
  }

  private boolean rotationalSymmetry180(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    Dimension blockDim = new Dimension(colorBlock.length, colorBlock[0].length);

    for (int i = 0; i < blockDim.width; i++) {
      for (int j = 0; j < blockDim.height; j++) {
        if (colorBlock[blockDim.width - 1 - i][blockDim.height - 1 - j]
            != colorBlockCompare[i][j]) {
          return false;
        }
      }
    }
    return true;
  }

  private boolean simpleSymmetry(Color[][] colorBlock, Color[][] colorBlockCompare)
  {
    Dimension blockDim = new Dimension(colorBlock.length, colorBlock[0].length);

    for (int i = 0; i < blockDim.width; i++) {
      for (int j = 0; j < blockDim.height; j++) {
        if (colorBlock[i][j] != colorBlockCompare[i][j]) {
          return false;
        }
      }
    }
    return true;
  }

  private Color[][] makeColorBlock(Color[][] picture, Rectangle subBlock)
  {
    Color[][] colorSubBlock = new Color[subBlock.width][subBlock.height];

    for (int i = 0; i < subBlock.width; i++) {
      for (int j = 0; j < subBlock.height; j++) {
        colorSubBlock[i][j] = picture[subBlock.x + i][subBlock.y + j];
      }
    }
    return colorSubBlock;
  }

/*
  public static void main(String[] args)
  {
    LifeDisplay ld = new LifeDisplay();
    ld.test();
  }
*/

  public void test()
  {
    Color[][] picture = new Color[2][2];

    picture[0][0] = Color.blue;
    picture[1][0] = Color.red;
    picture[0][1] = Color.red;
    picture[1][1] = Color.red;

    System.out.println(xAxisSymmetry(picture, picture));
    System.out.println(yAxisSymmetry(picture, picture));
    System.out.println(posDiagonalSymmetry(picture, picture));
    System.out.println(negDiagonalSymmetry(picture, picture));
    System.out.println(rotationalSymmetry90(picture, picture));
    System.out.println(rotationalSymmetry180(picture, picture));

    try {
      System.in.read();
    }
    catch (Exception exc) {
      System.out.println(exc);
      System.exit(1);
    }
  }

  public void print(Graphics pg)
  {
    Font f = new Font("Monospaced", Font.PLAIN, 14);
    FontMetrics fm = getFontMetrics(f);
    pg.setFont(f);
    int x = 10;
    int y = 10;
    int dy = fm.getHeight();

    y += dy;
    pg.drawString("Temperature:   " + temperatureLabel.getText(), x, y);
    y += dy;
    pg.drawString("Harmony:       " + harmonyLabel.getText(), x, y);
    y += dy;
    pg.drawString("Life:          " + lifeLabel.getText(), x, y);
    y += dy;
    pg.drawString("Complexity:    " + complexityLabel.getText(), x, y);

  }

}
